import importlib
from typing import Callable, List, Optional, Union

from chocs.http.http_error import NotFoundError
from chocs.http.http_method import HttpMethod
from chocs.http.http_request import HttpRequest
from chocs.http.http_response import HttpResponse
from .middleware.application_middleware import ApplicationMiddleware
from .middleware.middleware import Middleware, MiddlewarePipeline
from .routing import Route, Router
from .serverless.wrapper import create_serverless_function


class Application:
    def __init__(self, *middleware: Union[Middleware, Callable], open_api_path: str = ""):
        self.parent: Optional[Application] = None
        self._middleware = MiddlewarePipeline()
        for item in middleware:
            self._middleware.append(item)

        self.namespace = ["/"]
        self.router = Router()
        self._loaded_modules: List[str] = []
        self._cached_middleware: Optional[MiddlewarePipeline] = None

    def _append_route(
        self,
        method: HttpMethod,
        route: Route,
        handler: Callable[[HttpRequest], HttpResponse],
    ):
        if self.parent:
            self.parent._append_route(method, route, handler)

        self.router.append(route, handler, method)

    def _create_route(self, route: str, attributes: dict) -> Route:
        base_uri = "".join(self.namespace[1:])
        return Route(base_uri + route, attributes)

    def get(self, route: str, **attributes) -> Callable:
        def _get(handler: Callable) -> Callable:
            r = self._create_route(route, attributes)
            self._append_route(HttpMethod.GET, r, handler)
            return create_serverless_function(handler, r, self._middleware)

        return _get

    def post(self, route: str, **attributes) -> Callable:
        def _post(handler: Callable) -> Callable:
            r = self._create_route(route, attributes)
            self._append_route(HttpMethod.POST, r, handler)
            return create_serverless_function(handler, r, self._middleware)

        return _post

    def put(self, route: str, **attributes) -> Callable:
        def _put(handler: Callable) -> Callable:
            r = self._create_route(route, attributes)
            self._append_route(HttpMethod.PUT, r, handler)
            return create_serverless_function(handler, r, self._middleware)

        return _put

    def patch(self, route: str, **attributes) -> Callable:
        def _patch(handler: Callable) -> Callable:
            r = self._create_route(route, attributes)
            self._append_route(HttpMethod.PATCH, r, handler)
            return create_serverless_function(handler, r, self._middleware)

        return _patch

    def delete(self, route: str, **attributes) -> Callable:
        def _delete(handler: Callable) -> Callable:
            r = self._create_route(route, attributes)
            self._append_route(HttpMethod.DELETE, r, handler)
            return create_serverless_function(handler, r, self._middleware)

        return _delete

    def head(self, route: str, **attributes) -> Callable:
        def _head(handler: Callable) -> Callable:
            r = self._create_route(route, attributes)
            self._append_route(HttpMethod.HEAD, r, handler)
            return create_serverless_function(handler, r, self._middleware)

        return _head

    def options(self, route: str, **attributes) -> Callable:
        def _options(handler: Callable) -> Callable:
            r = self._create_route(route, attributes)
            self._append_route(HttpMethod.OPTIONS, r, handler)
            return create_serverless_function(handler, r, self._middleware)

        return _options

    def any(self, route: str, **attributes) -> Callable:
        def _any(handler: Callable) -> Callable:
            r = self._create_route(route, attributes)
            self._append_route(HttpMethod.GET, r, handler)
            self._append_route(HttpMethod.POST, r, handler)
            self._append_route(HttpMethod.PUT, r, handler)
            self._append_route(HttpMethod.PATCH, r, handler)
            self._append_route(HttpMethod.DELETE, r, handler)
            self._append_route(HttpMethod.HEAD, r, handler)
            self._append_route(HttpMethod.OPTIONS, r, handler)

            return create_serverless_function(handler, r, self._middleware)

        return _any

    def group(self, base_route: str) -> "Application":
        self.namespace.append(base_route)
        return self

    def __enter__(self) -> "Application":
        child_app = Application()
        child_app._middleware = self._middleware
        child_app.namespace = self.namespace
        child_app.parent = self
        child_app._loaded_modules = self._loaded_modules

        return child_app

    def __exit__(self, *args) -> "Application":
        self.namespace.pop()
        return self

    def __call__(self, request: HttpRequest) -> HttpResponse:
        try:
            route, handler = self.router.match(request.path, request.method)
            request.path_parameters = route.parameters
            request.route = route
            request.attributes["__handler__"] = handler
        except NotFoundError:

            def _handler(req: HttpRequest) -> HttpResponse:
                raise NotFoundError()

            request.attributes["__handler__"] = _handler

        return self._application_middleware(request)

    def use(self, module: str) -> None:
        if module in self._loaded_modules:
            return

        importlib.import_module(module)
        self._loaded_modules.append(module)

    @property
    def _application_middleware(self) -> MiddlewarePipeline:
        if self._cached_middleware is None:
            middleware = MiddlewarePipeline(self._middleware.queue)
            middleware.append(ApplicationMiddleware(self.router))
            self._cached_middleware = middleware

        return self._cached_middleware


__all__ = ["Application"]



from io import BytesIO
from json import dumps as json_dumps
from typing import Dict, Optional, Union

from chocs.http.http_headers import HttpHeaders
from chocs.http.http_method import HttpMethod
from chocs.http.http_query_string import HttpQueryString
from chocs.http.http_request import HttpRequest
from chocs.http.http_response import HttpResponse
from .application import Application


class TestClient:
    def __init__(self, application: Application):
        self.application = application

    def get(
        self,
        path: str,
        body: Union[Optional[BytesIO], str] = None,
        json: Optional[Dict] = None,
        headers: Optional[Union[HttpHeaders, Dict[str, str]]] = None,
    ) -> HttpResponse:

        return self.application(self._create_request(HttpMethod.GET, path, body, json, headers))

    def post(
        self,
        path: str,
        body: Union[Optional[BytesIO], str] = None,
        json: Optional[Dict] = None,
        headers: Optional[Union[HttpHeaders, Dict[str, str]]] = None,
    ) -> HttpResponse:

        return self.application(self._create_request(HttpMethod.POST, path, body, json, headers))

    def patch(
        self,
        path: str,
        body: Union[Optional[BytesIO], str] = None,
        json: Optional[Dict] = None,
        headers: Optional[Union[HttpHeaders, Dict[str, str]]] = None,
    ) -> HttpResponse:

        return self.application(self._create_request(HttpMethod.PATCH, path, body, json, headers))

    def delete(
        self,
        path: str,
        body: Union[Optional[BytesIO], str] = None,
        json: Optional[Dict] = None,
        headers: Optional[Union[HttpHeaders, Dict[str, str]]] = None,
    ) -> HttpResponse:

        return self.application(self._create_request(HttpMethod.DELETE, path, body, json, headers))

    def put(
        self,
        path: str,
        body: Union[Optional[BytesIO], str] = None,
        json: Optional[Dict] = None,
        headers: Optional[Union[HttpHeaders, Dict[str, str]]] = None,
    ) -> HttpResponse:

        return self.application(self._create_request(HttpMethod.PUT, path, body, json, headers))

    def options(
        self,
        path: str,
        body: Union[Optional[BytesIO], str] = None,
        json: Optional[Dict] = None,
        headers: Optional[Union[HttpHeaders, Dict[str, str]]] = None,
    ) -> HttpResponse:

        return self.application(self._create_request(HttpMethod.OPTIONS, path, body, json, headers))

    def head(
        self,
        path: str,
        body: Union[Optional[BytesIO], str] = None,
        json: Optional[Dict] = None,
        headers: Optional[Union[HttpHeaders, Dict[str, str]]] = None,
    ) -> HttpResponse:

        return self.application(self._create_request(HttpMethod.HEAD, path, body, json, headers))

    @staticmethod
    def _create_request(
        method: HttpMethod,
        path: str,
        body: Union[Optional[BytesIO], str],
        json: Optional[Dict] = None,
        headers: Optional[Union[HttpHeaders, Dict[str, str]]] = None,
    ) -> HttpRequest:
        if headers is None:
            headers = {}

        if json:
            body = json_dumps(json)
            headers["Content-Type"] = "application/json"

        path_parts = path.split("?")
        uri = path_parts[0]
        query_string = HttpQueryString(path_parts[1] if len(path_parts) > 1 else "")

        return HttpRequest(
            method=method,
            path=uri,
            body=body,
            headers=headers,
            query_string=query_string,
        )



import re
from copy import copy
from typing import Any, Callable, Dict, List, Optional, Pattern, Tuple, Union

from chocs.http.http_error import NotFoundError
from chocs.http.http_method import HttpMethod
from chocs.http.http_query_string import parse_qs_value

_ROUTE_REGEX = r"\\\{\s*(?P<var>[a-z_][a-z0-9_-]*)\s*\\\}"
_VAR_REGEX = "[^/]+"


class Route:
    __slots__ = [
        "route",
        "attributes",
        "_parameters_names",
        "_pattern",
        "_parameters",
        "is_wildcard",
    ]

    def __init__(self, route: str, attributes: Optional[Dict] = None):
        self.route = route
        self.attributes = attributes if attributes is not None else {}
        self._parameters_names: List[str] = []
        self._pattern: Pattern[str] = None  # type: ignore
        self._parameters: Dict[str, str] = {}
        self.is_wildcard: bool = "*" in route

    @property
    def pattern(self) -> Pattern[str]:
        if not self._pattern:
            self._parse()
        return self._pattern

    def _parse(self) -> None:
        def _parse_var(match):
            param = match.group(1)
            self._parameters_names.append(param)
            return f"({_VAR_REGEX})"

        pattern = re.escape(self.route)
        pattern = re.sub(_ROUTE_REGEX, _parse_var, pattern, re.I | re.M)
        pattern = re.sub(r"\\\*", ".*?", pattern)
        self._pattern = re.compile(
            "^" + pattern + "$",
            re.I | re.M,
        )

    def match(self, uri: str) -> Union[bool, "Route"]:
        matches = self.pattern.findall(uri)
        if not matches:
            return False

        if not self._parameters_names:
            return copy(self)

        if isinstance(matches[0], tuple):
            matches = list(matches[0])

        route = copy(self)
        match_index = 0
        for value in matches:
            route._parameters[self._parameters_names[match_index]] = parse_qs_value(value)
            match_index += 1

        return route

    @property
    def parameters(self):
        return self._parameters

    def __str__(self) -> str:
        return self.route

    def __bool__(self) -> bool:
        return True

    def __getitem__(self, key: str) -> str:
        return self._parameters[key]

    def __contains__(self, key: str) -> bool:
        return key in self._parameters

    def get(self, key: str, default: Optional[Any] = None):
        if key in self:
            return self[key]

        return default


class Router:
    def __init__(self):
        self._routes: Dict[HttpMethod, List[Tuple[Route, Callable]]] = {}

    def append(
        self,
        route: Route,
        handler: Callable,
        methods: Union[str, HttpMethod, List[Union[str, HttpMethod]]] = HttpMethod.GET,
    ) -> None:
        assert isinstance(route, Route), "Passed route must be instance of Route"
        normalised_methods = self._normalise_methods(methods)

        for method in normalised_methods:
            if method not in self._routes:
                self._routes[method] = []

            self._routes[method].append((route, handler))
            self._routes[method].sort(key=lambda r: r[0].is_wildcard)

    @staticmethod
    def _normalise_methods(methods: Union[str, HttpMethod, List[Union[str, HttpMethod]]]) -> List[HttpMethod]:
        if methods == "*":
            methods = [method for method in HttpMethod]
        elif isinstance(methods, HttpMethod):
            methods = [methods]
        else:
            methods = [method if isinstance(method, HttpMethod) else HttpMethod(method.upper()) for method in methods]

        return methods  # type: ignore

    def match(self, uri: str, method: Union[HttpMethod, str] = HttpMethod.GET) -> Tuple[Route, Callable]:
        if isinstance(method, str):
            method = HttpMethod(method)

        if method not in self._routes:
            raise NotFoundError(f"Could not match any resource matching {method} {uri} uri")

        for route in self._routes[method]:
            if route[0].match(uri):
                return route

        raise NotFoundError(f"Could not match any resource matching {method} {uri} uri")


__all__ = ["Route", "Router"]



from typing import List, Dict, Any

from .expression import parse_expression, Expression
from .sorting import parse_sorting

_RESERVED_FIELDS = ("sort", "limit", "order", "cursor", "offset")


class QueryCriteria:
    def __init__(self, query: Dict[str, Any], allowed_fields: List[str] = None):
        self._allowed_fields = allowed_fields
        self.limit = int(query["limit"]) if "limit" in query else 10

        if "cursor" in query:
            self.cursor = query["cursor"]
        else:
            self.offset = int(query["offset"]) if "offset" in query else 0

        self.sort = parse_sorting(query["sort"], allowed_fields) if "sort" in query else {}
        self._fields = create_criteria_fields(query, allowed_fields)

    def items(self):
        return self._fields.items()

    def keys(self):
        return self._fields.keys()

    def values(self):
        return self._fields.values()

    def next_query(self, cursor: str = "") -> str:
        query = [self._base_str]
        if cursor:
            query.append(f"cursor={cursor}")
        else:
            query.append(f"offset={self.offset + self.limit}")

        return "&".join(query)

    def prev_query(self, cursor: str = "") -> str:
        query = [self._base_str]
        if cursor:
            query.append(f"cursor={cursor}")
        elif self.offset - self.limit >= 0:
            query.append(f"offset={self.offset - self.limit}")
        else:
            query.append(f"offset=0")

        return "&".join(query)

    def __contains__(self, name: str) -> bool:
        return self._fields.__contains__(name)

    def __getitem__(self, key: str) -> Expression:
        return self._fields[key]

    def __iter__(self):
        return iter(self.items())

    @property
    def _base_str(self) -> str:
        result = []
        if self._fields:
            result.append("&".join([f"{field}={value}" for field, value in self._fields.items()]))

        if self.sort:
            result.append(",".join([f"{order}{name}" for name, order in self.sort.items()]))

        result.append(f"limit={self.limit}")

        return "&".join(result)

    def __str__(self) -> str:
        query = [self._base_str]

        if hasattr(self, "cursor"):
            query.append(f"cursor={self.cursor}")
        else:
            query.append(f"offset={self.offset}")

        return "&".join(query)

    def __repr__(self):
        return self.__str__()


def create_criteria_fields(query: Dict[str, Any], allowed_fields: List[str] = None) -> Dict[str, Expression]:
    result = {}

    for field, value in query.items():
        if field in _RESERVED_FIELDS:
            continue
        if allowed_fields and field not in allowed_fields:
            continue

        result[field] = parse_expression(value)

    return result


__all__ = ["QueryCriteria", "create_criteria_fields"]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['chocs',
 'chocs.http',
 'chocs.middleware',
 'chocs.query',
 'chocs.serverless',
 'chocs.wsgi']

package_data = \
{'': ['*']}

install_requires = \
['pyyaml>=5.3.1,<6.0.0']

setup_kwargs = {
    'name': 'chocs',
    'version': '1.1.0',
    'description': 'Lightweight and powerful WSGI/AWS lambda framework for rapid building rest applications.',
    'long_description': '<img align="right" width=300 src="https://github.com/kodemore/chocs/raw/master/chocs.png">\n\n# Chocs <br>[![PyPI version](https://badge.fury.io/py/chocs.svg)](https://pypi.org/project/chocs/) ![Release](https://github.com/kodemore/chocs/workflows/Release/badge.svg) ![Linting and Tests](https://github.com/kodemore/chocs/workflows/Linting%20and%20Tests/badge.svg) [![codecov](https://codecov.io/gh/kodemore/chocs/branch/master/graph/badge.svg)](https://codecov.io/gh/kodemore/chocs) [![Maintainability](https://api.codeclimate.com/v1/badges/9e3c979283b2361a9174/maintainability)](https://codeclimate.com/github/kodemore/chocs/maintainability)\n\nChocs is a modern HTTP framework for building AWS HTTP API/REST API and WSGI compatible applications. \nChocs aims to be small, expressive, and robust. \nIt provides an elegant API for writing fault-proof, extensible microservices.  \n\n \n## Features\n\n - AWS Serverless integration\n - Elegant and easy API\n - No additional bloat like built-in template engines, session handlers, etc.\n - Compatible with all WSGI servers\n - Loosely coupled components which can be used separately\n - Multipart body parsing\n - Graceful error handling\n - HTTP middleware support\n - Fast routing\n - Middleware packages to simplify daily tasks\n\n## Installation\n```\npip install chocs\n```\n\nor with poetry\n\n```\npoetry add chocs\n```\n\n## Quick start\n\n```python\nimport chocs\n\nhttp = chocs.Application()\n\n@http.get("/hello/{name}")\ndef hello(request: chocs.HttpRequest) -> chocs.HttpResponse:\n    return chocs.HttpResponse(f"Hello {request.path_parameters.get(\'name\')}!")\n\nchocs.serve(http)\n```\n\n> Keep in mind that the `chocs.serve()` function is using the `bjoern` package, so make sure you included it in your project\n> dependencies before using it. You are able to use any WSGI compatible server.\n\n## Available middlewares\n\n### OpenAPI Integration middleware\n\nAllows integrating OpenAPI documentation into your codebase, providing automating request validation based\non your OpenAPI spec. More details are available in the [chocs-openapi repository](https://github.com/kodemore/chocs-openapi).\n\n### ParsedBody middleware\n\nParsed body middleware helps to convert json/yaml request payloads into dataclass, this not only makes your\ndaily tasks easier but increases readability of your code and contract. More details are available in the [chocs-parsed-body repository](https://github.com/kodemore/chocs-parsed-body).\n\n\n# Documentation\nFor usage and detailed documentation please visit our [wiki page](https://github.com/kodemore/chocs/wiki)\n',
    'author': 'Dawid Kraczkowski',
    'author_email': 'dawid.kraczkowski@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kodemore/chocs',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from abc import ABC
from typing import Union, List

ExpressionValue = Union[str, int, float, None]


def _cast_value(value: str) -> ExpressionValue:
    if not value:
        return None

    if value.isnumeric():
        return int(value)

    if value.replace(".", "", 1).isdigit():
        return float(value)

    return value


class Expression(ABC):
    ...


class EqualExpression(Expression):
    value: ExpressionValue

    def __init__(self, value: ExpressionValue):
        self.value = value

    def __str__(self) -> str:
        return f"{self.value}"


class GreaterThanExpression(Expression):
    value: ExpressionValue

    def __init__(self, value: ExpressionValue):
        self.value = value

    def __str__(self) -> str:
        return f">{self.value}"


class LowerThanExpression(Expression):
    value: ExpressionValue

    def __init__(self, value: ExpressionValue):
        self.value = value

    def __str__(self) -> str:
        return f"<{self.value}"


class RangeExpression(Expression):
    start: ExpressionValue
    end: ExpressionValue

    def __init__(self, start: ExpressionValue, end: ExpressionValue):
        self.start = start
        self.end = end

    def __str__(self) -> str:
        return f"{self.start}..{self.end}"


class PatternExpression(Expression):
    start: ExpressionValue
    end: ExpressionValue

    def __init__(self, start: str, end: str):
        self.start = start
        self.end = end

    def __str__(self) -> str:
        return f"{self.start}*{self.end}"


class InExpression(Expression):
    values: List[ExpressionValue]

    def __init__(self, values: List[ExpressionValue]):
        self.values = values

    def __str__(self) -> str:
        return ",".join([str(value) for value in self.values])


def parse_expression(value: Union[list, str, int, float, bool]) -> Expression:
    if isinstance(value, list):
        return InExpression(value)

    if not isinstance(value, str):
        return EqualExpression(value)

    # edge cases
    if value in ("..", ">", "<", "*", ",", '"'):
        return EqualExpression(value)

    # escaping
    if value.startswith('"'):
        return EqualExpression(value[1:-1])

    if value.startswith(">"):
        return GreaterThanExpression(_cast_value(value[1:]))

    if value.startswith("<"):
        return LowerThanExpression(_cast_value(value[1:]))

    if ".." in value:
        range_values = value.split("..")
        if len(range_values) != 2:
            return EqualExpression(value)
        return RangeExpression(_cast_value(range_values[0]), _cast_value(range_values[1]))

    if "*" in value:
        pattern_values = value.split("*")
        if len(pattern_values) != 2:
            return EqualExpression(value)
        return PatternExpression(pattern_values[0], pattern_values[1])

    if "," in value:
        in_values = value.split(",")
        return InExpression([_cast_value(value.strip()) for value in in_values])

    # fallback to equal
    return EqualExpression(value)
